"use strict";
var __extends = (this && this.__extends) || (function () {
    var extendStatics = Object.setPrototypeOf ||
        ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
        function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
Object.defineProperty(exports, "__esModule", { value: true });
/*
 * Copyright 2019 Quadrant LLC. (http://quadrant.agency/)
 *
 * This file is part of Quadrant Transit.
 *
 * Quadrant Transit is distributed without license for evaluation purposes only.
 */
var EntityType;
(function (EntityType) {
    EntityType[EntityType["Table"] = 0] = "Table";
    EntityType[EntityType["View"] = 1] = "View";
    EntityType[EntityType["Procedure"] = 2] = "Procedure";
    EntityType[EntityType["Function"] = 3] = "Function";
})(EntityType = exports.EntityType || (exports.EntityType = {}));
var DataType;
(function (DataType) {
    DataType[DataType["Null"] = 0] = "Null";
    DataType[DataType["Boolean"] = 1] = "Boolean";
    DataType[DataType["Integer"] = 2] = "Integer";
    DataType[DataType["Float"] = 3] = "Float";
    DataType[DataType["String"] = 4] = "String";
    DataType[DataType["Date"] = 5] = "Date";
    DataType[DataType["DateTime"] = 6] = "DateTime";
    DataType[DataType["Array"] = 7] = "Array";
    DataType[DataType["Object"] = 8] = "Object";
    DataType[DataType["Any"] = 9] = "Any";
})(DataType = exports.DataType || (exports.DataType = {}));
var Command = /** @class */ (function () {
    function Command(spec, args) {
        var _this = this;
        this.spec = spec;
        this.args = spec.operation.parameters.map(function (param, i) {
            var arg = _this.validateArgument(param, args[i]);
            if (arg === undefined) {
                throw new Error("Parameter " + param.spec.name + " was not specified");
            }
            else {
                return arg;
            }
        });
    }
    Command.prototype.validateArgument = function (param, inarg) {
        var _this = this;
        var error = function (type) {
            throw new Error("Argument \"" + inarg + "\" for parameter \"" + param.spec.name + "\" of entity \"" + _this.spec.entity.name + "\" is not of data type \"" + type + "\".");
        };
        if (inarg == null && param.required) {
            throw new Error("Parameter " + param.spec.name + " of " + this.spec.entity.name + "." + this.spec.operation.name + " is required.");
        }
        // validate data types
        if ((inarg != undefined && inarg != null) || param.required) {
            var num = void 0, d = void 0;
            switch (param.spec.dataType) {
                case DataType.Null:
                    if (typeof inarg != "string" && inarg != "" && inarg.toLowerCase() != "null")
                        error("null");
                    break;
                case DataType.Boolean:
                    if (typeof inarg != "boolean")
                        error("boolean");
                    break;
                case DataType.Integer:
                    if (typeof inarg != "number" || !Number.isInteger(inarg))
                        error("integer");
                    break;
                case DataType.Float:
                    if (typeof inarg != "number" || isNaN(inarg))
                        error("float");
                    break;
                case DataType.String:
                    if (typeof inarg != "string")
                        error("string");
                    break;
                case DataType.Date:
                    num = Date.parse(inarg);
                    d = new Date(num);
                    if (isNaN(num) || d.getUTCHours() != 0 || d.getUTCMinutes() != 0 || d.getUTCSeconds() != 0 || d.getUTCMilliseconds() != 0)
                        error("date");
                    break;
                case DataType.DateTime:
                    num = Date.parse(inarg);
                    if (typeof inarg != "string" || (isNaN(num) && !inarg.match(/^\d\d:\d\d:\d\d$/)))
                        error("date/time");
                    break;
                case DataType.Array:
                    if (inarg == null)
                        inarg = [];
                    if (typeof inarg != "object" || !Array.isArray(inarg))
                        error("array");
                    break;
                case DataType.Object:
                    if (typeof inarg != "object" || Array.isArray(inarg))
                        error("object");
                    break;
                case DataType.Any:
                    break;
                default:
                    throw new Error("internal error: invalid data type: " + param.spec.dataType);
            }
        }
        // now translate defaults, if necessary
        // be careful, inarg may be undefined...
        var outarg = inarg;
        switch (param.spec.dataType) {
            case DataType.Null:
                outarg = null;
                break;
            case DataType.Boolean:
            case DataType.Integer:
            case DataType.Float:
            case DataType.String:
            case DataType.Date:
            case DataType.DateTime:
                break;
            case DataType.Array:
                if (inarg == undefined)
                    inarg = [];
                outarg = [];
                for (var _i = 0, inarg_1 = inarg; _i < inarg_1.length; _i++) {
                    var element = inarg_1[_i];
                    var elemArg = this.validateArgument(param.subparams[0], element);
                    if (elemArg == null) {
                        // can this even happen?
                        throw new Error("Array element value was not specified");
                    }
                    outarg.push(elemArg);
                }
                break;
            case DataType.Object:
                if (param.subparams.length == 0) {
                    outarg = inarg || {};
                }
                else {
                    outarg = {};
                    for (var _a = 0, _b = param.subparams; _a < _b.length; _a++) {
                        var subparam = _b[_a];
                        var subarg = this.validateArgument(subparam, inarg[subparam.spec.name]);
                        if (subarg !== undefined) {
                            outarg[subparam.spec.name] = subarg;
                        }
                    }
                    for (var name_1 in inarg) {
                        if (!(name_1 in outarg) && inarg[name_1] !== undefined) {
                            throw new Error("Unknown object key \"" + name_1 + "\" for parameter \"" + param.spec.name + "\" of entity \"" + this.spec.entity.name + "\"");
                        }
                    }
                }
                break;
            case DataType.Any:
                if (outarg === undefined)
                    outarg = null;
                break;
            default:
                throw new Error("internal error: invalid data type: " + param.spec.dataType);
        }
        return outarg;
    };
    Command.normalizeName = function (name, env) {
        name = name.replace(/^[^\w-$]+/, "").replace(/[^\w-$]+$/, "").replace(/[^\w-$]+/g, "_");
        if (env[name]) {
            var i = 2;
            while (env[name + i])
                i++;
            name += i;
        }
        return name;
    };
    return Command;
}());
exports.Command = Command;
var EntityBuilder = /** @class */ (function () {
    function EntityBuilder() {
        var _this = this;
        this.promise = new Promise(function (resolve, reject) {
            _this.resolve = resolve;
            _this.reject = reject;
        });
        this.entities = {};
    }
    EntityBuilder.prototype.entity = function (name, type, impl) {
        this.entities[name] = this.currentEntity = this.entities[name] || {
            name: name,
            type: type,
            impl: impl,
            identity: [],
            operations: {}
        };
        return this;
    };
    EntityBuilder.prototype.operation = function (name) {
        this.currentEntity.operations[name] = this.currentOperation = this.currentEntity.operations[name] || {
            name: name,
            parameters: [],
            outputs: []
        };
        return this;
    };
    EntityBuilder.prototype.identity = function (name, dataType, implType) {
        this.currentEntity.identity.push({
            name: name,
            dataType: dataType,
            implType: implType
        });
    };
    EntityBuilder.prototype.parameter = function (name, dataType, implType, required) {
        if (!this.currentOperation.parameters.find(function (op) { return op.spec.name == name; })) {
            this.currentOperation.parameters.push(this.currentParameter = {
                spec: {
                    name: name,
                    dataType: dataType,
                    implType: implType
                },
                required: required,
                subparams: []
            });
        }
        this.currentParameter = this.currentOperation.parameters.find(function (op) { return op.spec.name == name; });
        return this;
    };
    EntityBuilder.prototype.subparam = function (name, dataType, implType, required) {
        if (!this.currentParameter.subparams.find(function (param) { return param.spec.name == name; })) {
            this.currentParameter.subparams.push({
                spec: {
                    name: name,
                    dataType: dataType,
                    implType: implType
                },
                required: required,
                subparams: []
            });
        }
        return this;
    };
    EntityBuilder.prototype.output = function (name, dataType, implType, writeable) {
        this.currentOperation.outputs.push({
            spec: {
                name: name,
                dataType: dataType,
                implType: implType
            },
            writeable: writeable
        });
        return this;
    };
    EntityBuilder.prototype.done = function () {
        this.resolve(Object["values"](this.entities));
    };
    return EntityBuilder;
}());
exports.EntityBuilder = EntityBuilder;
var DatabaseEntityBuilder = /** @class */ (function (_super) {
    __extends(DatabaseEntityBuilder, _super);
    function DatabaseEntityBuilder() {
        return _super !== null && _super.apply(this, arguments) || this;
    }
    // for procs that take no params and return nothing
    DatabaseEntityBuilder.prototype.proc = function () {
        this.operation("call");
    };
    DatabaseEntityBuilder.prototype.procParameter = function (name, dataType, implType, required) {
        this.operation("call").parameter(name, dataType, implType, required);
    };
    DatabaseEntityBuilder.prototype.procOutput = function (name, dataType, implType) {
        this.operation("call").output(name, dataType, implType, false);
    };
    DatabaseEntityBuilder.prototype.tableColumn = function (name, dataType, implType, required, primaryKey, writeable) {
        var _this = this;
        this.column(name, dataType, implType, writeable);
        this.operation("truncate");
        if (primaryKey) {
            this.identity(name, dataType, implType);
            this.operation("insert").parameter(name, dataType, implType, writeable);
            this.operation("update").parameter(name, dataType, implType, true);
            this.operation("delete").parameter(name, dataType, implType, true);
        }
        else {
            this.operation("insert")
                .parameter("values", DataType.Object, "object", true)
                .subparam(name, dataType, implType, required && writeable);
            // we can insert w/o primary key, but we can't update
            if (this.currentEntity.identity.length) {
                if (this.currentEntity.identity.length == this.currentEntity.operations.update.parameters.length) {
                    // it's possible to change pks on update, so add them as values...
                    this.currentEntity.identity.forEach(function (pk) {
                        _this.operation("update").parameter("values", DataType.Object, "object", true)
                            .subparam(pk.name, pk.dataType, pk.implType, false);
                    });
                }
                this.operation("update")
                    .parameter("values", DataType.Object, "object", true)
                    .subparam(name, dataType, implType, required && writeable);
            }
        }
        return this;
    };
    DatabaseEntityBuilder.prototype.viewColumn = function (name, dataType, implType) {
        this.column(name, dataType, implType, false);
        return this;
    };
    DatabaseEntityBuilder.prototype.column = function (name, dataType, implType, writeable) {
        this.operation("select")
            .parameter("columns", DataType.Array, "array", false)
            .subparam("column", DataType.String, "string", false)
            .parameter("filter", DataType.Any, "any", false)
            .parameter("sort", DataType.Array, "array", false)
            .subparam("columns", DataType.String, "string", false);
        this.operation("select").output(name, dataType, implType, writeable);
        return this;
    };
    return DatabaseEntityBuilder;
}(EntityBuilder));
exports.DatabaseEntityBuilder = DatabaseEntityBuilder;
