"use strict";
/*
 * Copyright 2019 Quadrant LLC. (http://quadrant.agency/)
 *
 * This file is part of Quadrant Transit.
 *
 * Quadrant Transit is distributed without license for evaluation purposes only.
 */
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __generator = (this && this.__generator) || function (thisArg, body) {
    var _ = { label: 0, sent: function() { if (t[0] & 1) throw t[1]; return t[1]; }, trys: [], ops: [] }, f, y, t, g;
    return g = { next: verb(0), "throw": verb(1), "return": verb(2) }, typeof Symbol === "function" && (g[Symbol.iterator] = function() { return this; }), g;
    function verb(n) { return function (v) { return step([n, v]); }; }
    function step(op) {
        if (f) throw new TypeError("Generator is already executing.");
        while (_) try {
            if (f = 1, y && (t = op[0] & 2 ? y["return"] : op[0] ? y["throw"] || ((t = y["return"]) && t.call(y), 0) : y.next) && !(t = t.call(y, op[1])).done) return t;
            if (y = 0, t) op = [op[0] & 2, t.value];
            switch (op[0]) {
                case 0: case 1: t = op; break;
                case 4: _.label++; return { value: op[1], done: false };
                case 5: _.label++; y = op[1]; op = [0]; continue;
                case 7: op = _.ops.pop(); _.trys.pop(); continue;
                default:
                    if (!(t = _.trys, t = t.length > 0 && t[t.length - 1]) && (op[0] === 6 || op[0] === 2)) { _ = 0; continue; }
                    if (op[0] === 3 && (!t || (op[1] > t[0] && op[1] < t[3]))) { _.label = op[1]; break; }
                    if (op[0] === 6 && _.label < t[1]) { _.label = t[1]; t = op; break; }
                    if (t && _.label < t[2]) { _.label = t[2]; _.ops.push(op); break; }
                    if (t[2]) _.ops.pop();
                    _.trys.pop(); continue;
            }
            op = body.call(thisArg, _);
        } catch (e) { op = [6, e]; y = 0; } finally { f = t = 0; }
        if (op[0] & 5) throw op[1]; return { value: op[0] ? op[1] : void 0, done: true };
    }
};
Object.defineProperty(exports, "__esModule", { value: true });
var Command_1 = require("./Command");
var Result_1 = require("./Result");
var HandlerType;
(function (HandlerType) {
    HandlerType[HandlerType["Database"] = 0] = "Database";
    HandlerType[HandlerType["Declared"] = 1] = "Declared";
})(HandlerType = exports.HandlerType || (exports.HandlerType = {}));
var DataSourceFinder = /** @class */ (function () {
    function DataSourceFinder() {
        this.sources = {};
    }
    DataSourceFinder.prototype.addSource = function (dsconfig, processor, batchProcessor) {
        return __awaiter(this, void 0, void 0, function () {
            var ds, e_1, msg;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        _a.trys.push([0, 2, , 3]);
                        ds = new DataSource(dsconfig);
                        return [4 /*yield*/, ds.constructAPI(dsconfig.name, processor, batchProcessor)];
                    case 1:
                        _a.sent();
                        this.sources[dsconfig.name] = ds;
                        return [3 /*break*/, 3];
                    case 2:
                        e_1 = _a.sent();
                        msg = "Error configuring " + dsconfig.name + " DataSource with " + dsconfig.driver + " driver: " + e_1 + "\n";
                        if (e_1 instanceof Error) {
                            e_1.message = msg + e_1.message;
                            throw e_1;
                        }
                        else {
                            throw msg + e_1;
                        }
                        return [3 /*break*/, 3];
                    case 3: return [2 /*return*/];
                }
            });
        });
    };
    DataSourceFinder.prototype.hasSourceNamed = function (name) {
        return this.sources[name] != null;
    };
    DataSourceFinder.prototype.findSourceWithName = function (name) {
        if (this.sources[name]) {
            return this.sources[name];
        }
        else {
            throw new Error("Data source " + name + " not found.");
        }
    };
    return DataSourceFinder;
}());
exports.DataSourceFinder = DataSourceFinder;
var DataSource = /** @class */ (function () {
    function DataSource(dsconfig) {
        this._entities = {};
        this.handlers = {};
        this.extensions = [];
        this.config = dsconfig;
        var driverPath = dsconfig.driver.match(/^[\w\d]+$/) ? __dirname + "/driver/" + dsconfig.driver : dsconfig.driver;
        if (!dsconfig.name.match(/^\w[\w0-9_]*$/)) {
            throw new Error("Namespace must begin with a letter and may contain only letters, numbers, or underscores.");
        }
        this.handler = this.constructHandler(driverPath, dsconfig.driverConfig);
        for (var _i = 0, _a = this.config.extensions || []; _i < _a.length; _i++) {
            var path = _a[_i];
            this.extensions.push(this.constructHandler(path, {}));
        }
    }
    Object.defineProperty(DataSource.prototype, "hierarchy", {
        get: function () { return this._entities; },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(DataSource.prototype, "api", {
        get: function () { return this._api; },
        enumerable: true,
        configurable: true
    });
    DataSource.prototype.constructHandler = function (driverPath, driverConfig) {
        var driver = require.main.require(driverPath).createDriver(driverConfig);
        switch (driver.handlerType) {
            case HandlerType.Database: return new DatabaseHandler(driver);
            case HandlerType.Declared: return new DeclaredHandler(driver);
            default:
                throw new Error("The " + driverPath + " driver provided an unknown handler type: " + driver.handlerType);
        }
    };
    DataSource.prototype.getName = function () {
        return this.config.name;
    };
    DataSource.prototype.allowIdentity = function (ident) {
        if (this.config.allowedUsernames
            && ident.username
            && (this.config.allowedUsernames.find(function (val) { return val == ident.username; }) || this.config.allowedUsernames.find(function (val) { return val == "*"; }))) {
            return true;
        }
        if (this.config.allowedRoles) {
            var _loop_1 = function (role) {
                if (this_1.config.allowedRoles.find(function (val) { return val == role; }) || this_1.config.allowedRoles.find(function (val) { return val == "*"; })) {
                    return { value: true };
                }
            };
            var this_1 = this;
            for (var _i = 0, _a = ident.roles; _i < _a.length; _i++) {
                var role = _a[_i];
                var state_1 = _loop_1(role);
                if (typeof state_1 === "object")
                    return state_1.value;
            }
        }
        return false;
    };
    DataSource.prototype.hasEntity = function (entity) {
        try {
            this.specForEntity(entity);
            return true;
        }
        catch (e) {
            return false;
        }
    };
    DataSource.prototype.hasOperation = function (entity, operation) {
        try {
            this.specForOperation(entity, operation);
            return true;
        }
        catch (e) {
            return false;
        }
    };
    DataSource.prototype.defaultOperationName = function (entity) {
        var keys = Object.keys(this.specForEntity(entity).operations);
        if (keys.length == 1) {
            return keys[0];
        }
        else {
            return "";
        }
    };
    DataSource.prototype.specForEntity = function (entityName) {
        var entity = this._entities[entityName];
        if (!entity) {
            throw new Error("Entity " + this.config.name + "." + entityName + " not found");
        }
        return entity;
    };
    DataSource.prototype.handlerForEntity = function (entityName) {
        var source = this.handlers[entityName];
        if (!source) {
            throw new Error("Entity data source " + this.config.name + "." + entityName + " not found");
        }
        return source;
    };
    DataSource.prototype.specForOperation = function (entityName, operation) {
        var entity = this.specForEntity(entityName);
        var opSpec = entity.operations[operation];
        if (!opSpec) {
            throw new Error(operation + " operation not supported for entity " + this.config.name + "." + entityName);
        }
        return { entity: entity, operation: opSpec };
    };
    DataSource.prototype.constructAPI = function (namespace, processor, batchProcessor) {
        return __awaiter(this, void 0, void 0, function () {
            var entities, _i, entities_1, entity, apiName, _a, _b, ext, entities_3, _c, entities_2, entity, apiName, _d, _e, handler;
            return __generator(this, function (_f) {
                switch (_f.label) {
                    case 0: return [4 /*yield*/, this.handler.loadEntitySpecs()];
                    case 1:
                        entities = _f.sent();
                        for (_i = 0, entities_1 = entities; _i < entities_1.length; _i++) {
                            entity = entities_1[_i];
                            apiName = Command_1.Command.normalizeName(entity.name, this._entities);
                            this._entities[apiName] = entity;
                            this.handlers[apiName] = this.handler;
                        }
                        _a = 0, _b = this.extensions;
                        _f.label = 2;
                    case 2:
                        if (!(_a < _b.length)) return [3 /*break*/, 5];
                        ext = _b[_a];
                        return [4 /*yield*/, ext.loadEntitySpecs()];
                    case 3:
                        entities_3 = _f.sent();
                        for (_c = 0, entities_2 = entities_3; _c < entities_2.length; _c++) {
                            entity = entities_2[_c];
                            apiName = Command_1.Command.normalizeName(entity.name, {});
                            this._entities[apiName] = entity;
                            this.handlers[apiName] = ext;
                        }
                        _f.label = 4;
                    case 4:
                        _a++;
                        return [3 /*break*/, 2];
                    case 5:
                        this._api = constructAPIFromHierarchy(this._entities, namespace, processor, batchProcessor);
                        for (_d = 0, _e = Object["values"](this.handlers); _d < _e.length; _d++) {
                            handler = _e[_d];
                            handler.useAPI(this.api);
                        }
                        return [2 /*return*/];
                }
            });
        });
    };
    DataSource.prototype.executeCommand = function (entityName, operationName, args) {
        return __awaiter(this, void 0, void 0, function () {
            var spec, handler, command;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        spec = this.specForOperation(entityName, operationName);
                        handler = this.handlerForEntity(entityName);
                        if (spec == null) {
                            throw new Error("Object not found: " + this.config.name + "." + entityName);
                        }
                        command = new Command_1.Command(spec, args || []);
                        return [4 /*yield*/, handler.executeCommand(command)];
                    case 1: return [2 /*return*/, _a.sent()];
                }
            });
        });
    };
    DataSource.prototype.executeBatch = function (args) {
        return __awaiter(this, void 0, void 0, function () {
            var commands, i, _a, entityName, operationName, spec;
            return __generator(this, function (_b) {
                switch (_b.label) {
                    case 0:
                        commands = [];
                        for (i = 0; i < args.length; i++) {
                            _a = args[i][0].split("."), entityName = _a[0], operationName = _a[1];
                            spec = this.specForOperation(entityName, operationName);
                            if (this.handlerForEntity(entityName) != this.handler) {
                                throw new Error("Batches may not contain APIs provided by extensions");
                            }
                            if (spec == null) {
                                throw new Error("Object not found: " + this.config.name + "." + entityName);
                            }
                            commands.push(new Command_1.Command(spec, args[i][1] || []));
                        }
                        return [4 /*yield*/, this.handler.executeBatch(commands)];
                    case 1: return [2 /*return*/, _b.sent()];
                }
            });
        });
    };
    DataSource.prototype.executeSQL = function (sql) {
        return __awaiter(this, void 0, void 0, function () {
            return __generator(this, function (_a) {
                if (this.handler.handlerType == HandlerType.Database) {
                    return [2 /*return*/, this.handler.executeSQL(sql)];
                }
                else {
                    throw new Error("Cannot call executeSQL with driver " + this.config.driver);
                }
                return [2 /*return*/];
            });
        });
    };
    DataSource.prototype.getAPIJavaScriptConstructor = function () {
        return constructAPIFromHierarchy.toString();
    };
    return DataSource;
}());
exports.DataSource = DataSource;
var DatabaseHandler = /** @class */ (function () {
    function DatabaseHandler(driver) {
        this.handlerType = HandlerType.Database;
        this.driver = driver;
    }
    DatabaseHandler.prototype.acquireConnection = function () {
        return __awaiter(this, void 0, void 0, function () {
            var notifier;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        notifier = new Result_1.OperationNotifier();
                        this.driver.openConnection(notifier);
                        return [4 /*yield*/, notifier.promise];
                    case 1: return [2 /*return*/, _a.sent()];
                }
            });
        });
    };
    DatabaseHandler.prototype.releaseConnection = function (conn) {
        return __awaiter(this, void 0, void 0, function () {
            return __generator(this, function (_a) {
                this.driver.closeConnection(conn);
                return [2 /*return*/];
            });
        });
    };
    DatabaseHandler.prototype.loadEntitySpecs = function () {
        return __awaiter(this, void 0, void 0, function () {
            var conn, builder;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0: return [4 /*yield*/, this.acquireConnection()];
                    case 1:
                        conn = _a.sent();
                        _a.label = 2;
                    case 2:
                        _a.trys.push([2, , 5, 6]);
                        builder = new Command_1.DatabaseEntityBuilder();
                        return [4 /*yield*/, this.driver.loadEntitySpecs(conn, builder)];
                    case 3:
                        _a.sent();
                        return [4 /*yield*/, builder.promise];
                    case 4: return [2 /*return*/, _a.sent()];
                    case 5:
                        this.releaseConnection(conn);
                        return [7 /*endfinally*/];
                    case 6: return [2 /*return*/];
                }
            });
        });
    };
    DatabaseHandler.prototype.useAPI = function (api) {
    };
    DatabaseHandler.prototype.executeCommand = function (command) {
        return __awaiter(this, void 0, void 0, function () {
            var conn, notifier, response;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0: return [4 /*yield*/, this.acquireConnection()];
                    case 1:
                        conn = _a.sent();
                        _a.label = 2;
                    case 2:
                        _a.trys.push([2, , 4, 5]);
                        notifier = new Result_1.DatabaseResultNotifier();
                        this.driver.executeCommand(conn, command, notifier);
                        return [4 /*yield*/, notifier.promise];
                    case 3:
                        response = _a.sent();
                        return [2 /*return*/, {
                                data: response,
                                meta: { handler: this.handlerType }
                            }];
                    case 4:
                        this.releaseConnection(conn);
                        return [7 /*endfinally*/];
                    case 5: return [2 /*return*/];
                }
            });
        });
    };
    DatabaseHandler.prototype.executeBatch = function (commands) {
        return __awaiter(this, void 0, void 0, function () {
            var results, conn, index, notifier, _a, _b, err_1;
            return __generator(this, function (_c) {
                switch (_c.label) {
                    case 0:
                        results = [];
                        return [4 /*yield*/, this.acquireConnection()];
                    case 1:
                        conn = _c.sent();
                        index = 0;
                        _c.label = 2;
                    case 2:
                        _c.trys.push([2, 9, 11, 12]);
                        return [4 /*yield*/, this.driver.beginTransaction(conn)];
                    case 3:
                        _c.sent();
                        index = 0;
                        _c.label = 4;
                    case 4:
                        if (!(index < commands.length)) return [3 /*break*/, 7];
                        notifier = new Result_1.DatabaseResultNotifier();
                        this.driver.executeCommand(conn, commands[index], notifier);
                        _a = results;
                        _b = results.length;
                        return [4 /*yield*/, notifier.promise];
                    case 5:
                        _a[_b] = _c.sent();
                        _c.label = 6;
                    case 6:
                        index++;
                        return [3 /*break*/, 4];
                    case 7: return [4 /*yield*/, this.driver.commitTransaction(conn)];
                    case 8:
                        _c.sent();
                        return [3 /*break*/, 12];
                    case 9:
                        err_1 = _c.sent();
                        return [4 /*yield*/, this.driver.rollbackTransaction(conn)];
                    case 10:
                        _c.sent();
                        if (typeof err_1 != "object") {
                            err_1 = new Error(err_1);
                        }
                        err_1.meta = { batchIndex: index };
                        throw err_1;
                    case 11:
                        this.releaseConnection(conn);
                        return [7 /*endfinally*/];
                    case 12: return [2 /*return*/, {
                            data: results,
                            meta: { handler: this.handlerType }
                        }];
                }
            });
        });
    };
    DatabaseHandler.prototype.executeSQL = function (sql) {
        return __awaiter(this, void 0, void 0, function () {
            var conn, notifier;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0: return [4 /*yield*/, this.acquireConnection()];
                    case 1:
                        conn = _a.sent();
                        _a.label = 2;
                    case 2:
                        _a.trys.push([2, , 4, 5]);
                        notifier = new Result_1.DatabaseResultNotifier();
                        this.driver.executeSQL(conn, sql, notifier, null);
                        return [4 /*yield*/, notifier.promise];
                    case 3: return [2 /*return*/, _a.sent()];
                    case 4:
                        this.releaseConnection(conn);
                        return [7 /*endfinally*/];
                    case 5: return [2 /*return*/];
                }
            });
        });
    };
    return DatabaseHandler;
}());
var DeclaredHandler = /** @class */ (function () {
    function DeclaredHandler(driver) {
        this.handlerType = HandlerType.Declared;
        this.driver = driver;
    }
    DeclaredHandler.prototype.loadEntitySpecs = function () {
        return __awaiter(this, void 0, void 0, function () {
            var entities, _loop_2, this_2, consumed, remaining, spec;
            return __generator(this, function (_a) {
                entities = [];
                _loop_2 = function (spec) {
                    var regex = {
                        identifier: /^([a-zA-Z_]\w*)/
                    };
                    consumed = 0;
                    remaining = String(spec);
                    function error(msg) {
                        return new Error("Unable to parse API signature - " + msg + " at character " + consumed + ": " + spec);
                    }
                    function whitespace() {
                        if (remaining.match(/^(\s+)/)) {
                            consumed += RegExp.$1.length;
                            remaining = remaining.slice(RegExp.$1.length);
                        }
                    }
                    function peekString(atom) {
                        whitespace();
                        return remaining.indexOf(atom) == 0;
                    }
                    function expectString(atom) {
                        if (peekString(atom)) {
                            consumed += atom.length;
                            remaining = remaining.slice(atom.length);
                            return atom;
                        }
                        else {
                            throw error("expected: " + atom);
                        }
                    }
                    function peekMatch(re) {
                        whitespace();
                        return remaining.match(regex[re]);
                    }
                    function expectMatch(re) {
                        if (peekMatch(re)) {
                            consumed += RegExp.$1.length;
                            remaining = remaining.slice(RegExp.$1.length);
                            return RegExp.$1;
                        }
                        else {
                            throw error("expected " + re);
                        }
                    }
                    function getType(type) {
                        switch (type) {
                            case "string": return Command_1.DataType.String;
                            case "float": return Command_1.DataType.Float;
                            case "int": return Command_1.DataType.Integer;
                            case "boolean": return Command_1.DataType.Boolean;
                            case "null": return Command_1.DataType.Null;
                            default: throw error("unknown type " + type);
                        }
                    }
                    var entity = expectMatch("identifier");
                    expectString("(");
                    var params = [];
                    while (peekMatch("identifier")) {
                        var name_1 = expectMatch("identifier");
                        expectString(":");
                        var type_1 = expectMatch("identifier");
                        params.push({
                            spec: {
                                name: name_1,
                                dataType: getType(type_1)
                            },
                            required: true,
                            subparams: []
                        });
                        if (!peekString(",")) {
                            break;
                        }
                    }
                    expectString(")");
                    expectString(":");
                    var outputs = [];
                    var type = expectMatch("identifier");
                    outputs.push({
                        spec: {
                            name: "value",
                            dataType: getType(type)
                        },
                        writeable: false
                    });
                    entities.push({
                        name: entity,
                        type: Command_1.EntityType.Function,
                        identity: [],
                        impl: this_2.driver.API[spec],
                        operations: {
                            call: {
                                name: "call",
                                parameters: params,
                                outputs: outputs
                            }
                        }
                    });
                };
                this_2 = this;
                for (spec in this.driver.API) {
                    _loop_2(spec);
                }
                return [2 /*return*/, Promise.resolve(entities)];
            });
        });
    };
    DeclaredHandler.prototype.useAPI = function (api) {
        if (this.driver.useNamespaceAPI)
            this.driver.useNamespaceAPI(api);
    };
    DeclaredHandler.prototype.executeCommand = function (command) {
        return __awaiter(this, void 0, void 0, function () {
            var value;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0: return [4 /*yield*/, command.spec.entity.impl.apply(this.driver.API, command.args)];
                    case 1:
                        value = _a.sent();
                        return [2 /*return*/, Promise.resolve({
                                data: value,
                                meta: { handler: this.handlerType }
                            })];
                }
            });
        });
    };
    DeclaredHandler.prototype.executeBatch = function (commands) {
        return __awaiter(this, void 0, void 0, function () {
            var results, index, result, err_2;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        results = [];
                        index = 0;
                        _a.label = 1;
                    case 1:
                        if (!(index < commands.length)) return [3 /*break*/, 7];
                        result = void 0;
                        _a.label = 2;
                    case 2:
                        _a.trys.push([2, 4, , 5]);
                        return [4 /*yield*/, this.executeCommand(commands[index])];
                    case 3:
                        result = _a.sent();
                        return [3 /*break*/, 5];
                    case 4:
                        err_2 = _a.sent();
                        if (typeof err_2 != "object") {
                            err_2 = new Error(err_2);
                        }
                        err_2.meta = { batchIndex: index };
                        throw err_2;
                    case 5:
                        results[results.length] = result.data;
                        _a.label = 6;
                    case 6:
                        index++;
                        return [3 /*break*/, 1];
                    case 7: return [2 /*return*/, {
                            data: results,
                            meta: { handler: this.handlerType }
                        }];
                }
            });
        });
    };
    return DeclaredHandler;
}());
// we're going to toString this for use in the browser,
// so we need to use old-school JS syntax...
function constructAPIFromHierarchy(hier, ns, processor, batchProcessor) {
    function remapRows(result) {
        var rows = [];
        for (var i = 0; i < result.rows.length; i++) {
            var row = result.rows[i];
            var data = {};
            for (var j = 0; j < result.columns.length; j++) {
                data[result.columns[j].spec.name] = row[j];
            }
            rows.push(data);
        }
        rows["columns"] = result.columns;
        return rows;
    }
    var ret = {
        _: {
            TABLE: 0, VIEW: 1, PROCEDURE: 2, FUNCTION: 3,
            NULL: 0, BOOLEAN: 1, INTEGER: 2, FLOAT: 3, STRING: 4, DATE: 5, DATETIME: 6, ARRAY: 7, OBJECT: 8, ANY: 9,
        },
        "$batch_api": (function (d) {
            return function (args) {
                return new Promise(function (resolve, reject) {
                    batchProcessor(d, args)
                        .then(function (response) {
                        if (response.meta.handler == 0) {
                            var remapped = [];
                            for (var i = 0; i < response.data.length; i++) {
                                remapped[i] = remapRows(response.data[i]);
                            }
                            resolve(remapped);
                        }
                        else {
                            resolve(response.data);
                        }
                    })
                        .catch(function (err) {
                        reject(err);
                    });
                });
            };
        })(ns)
    };
    for (var entity in hier) {
        ret[entity] = {};
        for (var operation in hier[entity].operations) {
            (function (e, o) {
                ret[e][o] = function () {
                    var args = Array.prototype.slice.call(arguments);
                    return new Promise(function (resolve, reject) {
                        processor(ns, e, o, args)
                            .then(function (response) {
                            if (response.meta.handler == 0) {
                                resolve(remapRows(response.data));
                            }
                            else {
                                resolve(response.data);
                            }
                        })
                            .catch(function (err) {
                            reject(err);
                        });
                    });
                };
            })(entity, operation);
        }
        if (hier[entity].type == ret._.TABLE) {
            (function (entName, entSpec, apiOut) {
                function unflattenRows(flat) {
                    var headers = flat[0];
                    var rows = [];
                    var fieldSpecs = {};
                    var outputs = entSpec.operations.select.outputs;
                    for (var i = 0; i < outputs.length; i++) {
                        fieldSpecs[outputs[i].spec.name.toLowerCase()] = outputs[i].spec;
                    }
                    for (var i = 1; i < flat.length; i++) {
                        var row = {};
                        for (var j = 0; j < flat[i].length; j++) {
                            if (!headers[j])
                                continue;
                            var fieldSpec = fieldSpecs[String(headers[j]).toLowerCase()];
                            if (!fieldSpec) {
                                throw new Error("Unknown column: " + headers[j]);
                            }
                            var value = flat[i][j];
                            if (value != null) {
                                switch (fieldSpec.dataType) {
                                    case ret._.BOOLEAN:
                                        if (typeof (value) == "string") {
                                            value = value.toLowerCase();
                                        }
                                        if (value == "n" || value == "no" || value == "f" || value == "false" || value == "0") {
                                            value = false;
                                        }
                                        else {
                                            value = !!value;
                                        }
                                        break;
                                    case ret._.INTEGER:
                                    case ret._.FLOAT:
                                        value = Number(value);
                                        if (isNaN(value)) {
                                            throw new Error("Value for column " + fieldSpec.name + " in row " + i + " is not a number");
                                        }
                                        break;
                                    case ret._.STRING:
                                        value = (value == null) ? null : String(value);
                                        break;
                                    default:
                                    // FALL THROUGH
                                }
                                row[fieldSpec.name] = value;
                            }
                        }
                        rows.push(row);
                    }
                    return rows;
                }
                function buildRowBatch(rows) {
                    var batch = [];
                    for (var i = 0; i < rows.length; i++) {
                        var params = [];
                        var attrs = {};
                        for (var n in rows[i]) {
                            attrs[n] = rows[i][n];
                        }
                        for (var j = 0; j < entSpec.identity.length; j++) {
                            var name = entSpec.identity[j].name;
                            params.push(attrs[name]);
                            delete attrs[name];
                        }
                        params.push(attrs);
                        batch.push([entName + ".insert", params]);
                    }
                    return batch;
                }
                apiOut.buildInsertBatch = function (rows) {
                    return buildRowBatch(rows);
                },
                    apiOut.insertRows = function (rows) {
                        return ret.$batch_api(this.buildInsertBatch(rows));
                    };
                apiOut.insertRowsFlat = function (flat) {
                    return this.insertRows(unflattenRows(flat));
                };
                apiOut.buildReplaceBatch = function (rows) {
                    var batch = this.buildInsertBatch(rows);
                    batch.unshift([entName + ".truncate"]);
                    return batch;
                },
                    apiOut.replaceRows = function (rows) {
                        return ret.$batch_api(this.buildReplaceBatch(rows));
                    };
                apiOut.replaceRowsFlat = function (flat) {
                    return this.replaceRows(unflattenRows(flat));
                };
            })(entity, hier[entity], ret[entity]);
        }
    }
    return ret;
}
exports.constructAPIFromHierarchy = constructAPIFromHierarchy;
function keys(o) {
    var k = [];
    for (var n in o) {
        k.push(n);
    }
    k.sort();
    return k;
}
