"use strict";
/*
 * Copyright 2019 Quadrant LLC. (http://quadrant.agency/)
 *
 * This file is part of Quadrant Transit.
 *
 * Quadrant Transit is distributed without license for evaluation purposes only.
 */
Object.defineProperty(exports, "__esModule", { value: true });
var rotating_file_stream_1 = require("rotating-file-stream");
var logs = {
    event: null,
    api: null,
    doc: null,
};
var rotationInterval = "5s";
exports.enabled = true;
function makeLogger(name, file) {
    if (file) {
        var path = './';
        if (file.match(/^(.*)\/([^\/]+)$/)) {
            path = RegExp.$1;
            file = RegExp.$2;
        }
        var pad_1 = function (n) { return n < 10 ? '0' + n : n; };
        logs[name] = rotating_file_stream_1.default(function (t, i) { return file + (t ? "." + t.getFullYear() + "-" + pad_1(t.getMonth() + 1) + "-" + pad_1(t.getDate()) : ""); }, { path: path, interval: "1d" });
        logs[name].on("error", function (err) {
            console.log(new Date().toISOString() + " Error logging " + name + " to " + file + ", " + name + " logging disabled: " + err);
            logs[name] = null;
        });
    }
    else {
        logs[name] = null;
    }
}
function configureEvent(file) {
    makeLogger("event", file);
}
exports.configureEvent = configureEvent;
function configureAPI(file) {
    makeLogger("api", file);
}
exports.configureAPI = configureAPI;
function configureDocument(file) {
    makeLogger("doc", file);
}
exports.configureDocument = configureDocument;
function writeEventArgs(level, pargs) {
    if (exports.enabled) {
        var args = [new Date().toISOString(), level].concat(Array.prototype.slice.call(pargs));
        console.log.apply(console, args);
        if (logs.event) {
            logs.event.write(args.join(" "));
            logs.event.write("\n");
        }
    }
}
function error() {
    var any = [];
    for (var _i = 0; _i < arguments.length; _i++) {
        any[_i] = arguments[_i];
    }
    writeEventArgs("error", arguments);
}
exports.error = error;
function warn() {
    var any = [];
    for (var _i = 0; _i < arguments.length; _i++) {
        any[_i] = arguments[_i];
    }
    writeEventArgs("warn", arguments);
}
exports.warn = warn;
function info() {
    var any = [];
    for (var _i = 0; _i < arguments.length; _i++) {
        any[_i] = arguments[_i];
    }
    writeEventArgs("info", arguments);
}
exports.info = info;
function writeTransaction(logger, req, res) {
    if (exports.enabled) {
        if (logger) {
            logger.write(new Date().toISOString());
            logger.write(" ");
            logger.write(String(res.statusCode));
            logger.write(" ");
            logger.write(String(req.url));
            logger.write("\n");
        }
    }
}
function api(req, res) {
    writeTransaction(logs.api, req, res);
}
exports.api = api;
function document(req, res) {
    writeTransaction(logs.doc, req, res);
}
exports.document = document;
